from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.types import Course, File


class FolderMod(MoodleMod):
    MOD_NAME = 'folder'
    MOD_PLURAL_NAME = 'folders'
    MOD_MIN_VERSION = 2017051500  # 3.3

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        # TODO: Add download condition
        return True

    async def real_fetch_mod_entries(self, courses: List[Course]) -> Dict[int, Dict[int, Dict]]:
        folders = (
            await self.client.async_post(
                'mod_folder_get_folders_by_courses', self.get_data_for_mod_entries_endpoint(courses)
            )
        ).get('folders', [])

        result = {}
        for folder in folders:
            course_id = folder.get('course', 0)
            folder_files = folder.get('introfiles', [])
            folder_time_modified = folder.get('timemodified', 0)
            self.set_files_types_if_empty(folder_files, 'folder_file')

            folder_intro = folder.get('intro', '')
            if folder_intro != '':
                folder_files.append(
                    {
                        'filename': 'Folder intro',
                        'filepath': '/',
                        'description': folder_intro,
                        'timemodified': folder_time_modified,
                        'filter_urls_during_search_containing': ['/mod_folder/intro'],
                        'type': 'description',
                    }
                )

            self.add_module(
                result,
                course_id,
                folder.get('coursemodule', 0),
                {
                    'id': folder.get('id', 0),
                    'name': folder.get('name', 'unnamed folder'),
                    'timemodified': folder_time_modified,
                    'files': folder_files,
                },
            )

        return result



import sys

from getpass import getpass

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.moodle_service import MoodleService
from moodle_dl.moodle.request_helper import RequestRejectedError
from moodle_dl.types import MoodleURL, MoodleDlOpts
from moodle_dl.utils import Log


class MoodleWizard:
    def __init__(self, config: ConfigHelper, opts: MoodleDlOpts):
        self.config = config
        self.opts = opts

    def interactively_acquire_token(self, use_stored_url: bool = False) -> str:
        if self.opts.sso or self.opts.token is not None:
            self.interactively_acquire_sso_token(use_stored_url=use_stored_url)
        else:
            self.interactively_acquire_normal_token(use_stored_url=use_stored_url)

    def interactively_get_moodle_url(self, use_stored_url: bool) -> MoodleURL:
        if use_stored_url:
            return self.config.get_moodle_URL()

        url_ok = False
        while not url_ok:
            url_ok = True
            moodle_url = input('URL of Moodle:   ')

            use_http = False
            if moodle_url.startswith('http://'):
                Log.warning(
                    'Warning: You have entered an insecure URL! Are you sure that the Moodle is'
                    + ' not accessible via `https://`? All your data will be transferred'
                    + ' insecurely! If your Moodle is accessible via `https://`, then run'
                    + ' the process again using `https://` to protect your data.'
                )
                use_http = True
            elif not moodle_url.startswith('https://'):
                Log.error('The url of your moodle must start with `https://`')
                url_ok = False

        moodle_domain, moodle_path = MoodleService.split_moodle_url(moodle_url)
        return MoodleURL(use_http, moodle_domain, moodle_path)

    def interactively_acquire_normal_token(self, use_stored_url: bool = False) -> str:
        """
        Walks the user through executing a login into the Moodle-System to get
        the Token and saves it.
        @return: The Token for Moodle.
        """

        automated = False
        stop_automatic_generation = False
        if self.opts.username is not None and self.opts.password is not None:
            automated = True

        if not automated:
            print('[The following Credentials are not saved, it is only used temporarily to generate a login token.]')

        moodle_token = None
        while moodle_token is None or (stop_automatic_generation and automated):
            moodle_url = self.interactively_get_moodle_url(use_stored_url)

            if self.opts.username is not None:
                moodle_username = self.opts.username
                stop_automatic_generation = True
            else:
                moodle_username = input('Username for Moodle:   ')

            if self.opts.password is not None:
                moodle_password = self.opts.password
            else:
                moodle_password = getpass('Password for Moodle [no output]:   ')

            try:
                moodle_token, moodle_privatetoken = MoodleService.obtain_login_token(
                    self.opts, moodle_username, moodle_password, moodle_url
                )

            except RequestRejectedError as error:
                Log.error(f'Login Failed! ({error}) Please try again.')
            except (ValueError, RuntimeError) as error:
                Log.error(f'Error while communicating with the Moodle System! ({error}) Please try again.')
            except ConnectionError as error:
                Log.error(str(error))

        if automated is True and moodle_token is None:
            sys.exit(1)

        self.config.set_tokens(moodle_token, moodle_privatetoken)
        self.config.set_moodle_URL(moodle_url)

        Log.success('Token successfully saved!')

        return moodle_token

    def interactively_acquire_sso_token(self, use_stored_url: bool = False) -> str:
        """
        Walks the user through the receiving of a SSO token
        @return: The Token for Moodle.
        """

        moodle_url = self.interactively_get_moodle_url(use_stored_url)

        if self.opts.token is not None:
            moodle_token = self.opts.token
        else:
            Log.warning('Please use the Chrome browser for the following procedure')
            print('1. Log into your Moodle Account')
            print('2. Open the developer console (press F12) and go to the Network tab')
            print('3. Then visit the following URL in the same browser tab you have logged in:')

            print(
                moodle_url.url_base
                + 'admin/tool/mobile/launch.php?service=moodle_mobile_app&passport=12345&urlscheme=moodledl'
            )
            print()
            print(
                'If you open the link, no web page should load, instead an error will occur.'
                + ' In the Network tab of the developer console you opened before there should be an error entry.'
            )

            print('The script expects a URL that looks something like this:')
            Log.info('moodledl://token=$apptoken')
            print(
                ' Where $apptoken looks random and "moodledl" can also be a different url scheme '
                + ' like "moodlemobile". In reality $apptoken is a Base64 string containing the token to access moodle.'
            )

            print(
                '4. Copy the link address of the website that could not be loaded'
                + ' (right click the list entry, then click on Copy, then click on copy link address)'
            )

            token_address = input('Then insert the link address here:   ')

            moodle_token, moodle_privatetoken = MoodleService.extract_token(token_address)
            if moodle_token is None:
                raise ValueError('Invalid URL!')

        self.config.set_tokens(moodle_token, moodle_privatetoken)
        self.config.set_moodle_URL(moodle_url)

        Log.success('Token successfully saved!')

        return moodle_token



import collections
import contextlib
import email.utils
import getpass
import html
import http
import io
import itertools
import logging
import math
import os
import re
import shutil
import ssl
import sys
import time
import unicodedata

from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Dict
from aiohttp.cookiejar import CookieJar

import readchar
import requests
import urllib3

from requests.utils import DEFAULT_CA_BUNDLE_PATH, extract_zipped_paths


def check_verbose() -> bool:
    """Return if the verbose mode is active"""
    return '-v' in sys.argv or '--verbose' in sys.argv


def check_debug() -> bool:
    """Return if the debugger is currently active"""
    return 'pydevd' in sys.modules or (hasattr(sys, 'gettrace') and sys.gettrace() is not None)


def format_seconds(seconds):
    (mins, secs) = divmod(seconds, 60)
    (hours, mins) = divmod(mins, 60)
    if hours > 99:
        return '--:--:--'
    if hours == 0:
        return f'{int(mins):02d}:{int(secs):02d}'
    return f'{int(hours):02d}:{int(mins):02d}:{int(secs):02d}'


def calc_speed(start, now, byte_count):
    dif = now - start
    if byte_count <= 0 or dif < 0.001:  # One millisecond
        return None
    return float(byte_count) / dif


def format_speed(speed):
    if speed is None:
        return f"{'---b/s':10}"
    speed_text = format_bytes(speed) + '/s'
    return f'{speed_text:10}'


async def run_with_final_message(load_function, entry: Dict, message: str, *format_args):
    result = await load_function(entry)
    logging.info(message, *format_args)
    return result


def get_nested(from_dict: Dict, key: str, default=None):
    keys = key.split('.')
    try:
        result = from_dict
        for key in keys:
            result = result[key]
        return result
    except KeyError:
        return default


KNOWN_EXTENSIONS = (
    ['avi', 'flv', 'mkv', 'mov', 'mp4', 'webm', '3g2', '3gp', 'f4v', 'mk3d', 'divx', 'mpg', 'ogv', 'm4v']
    + ['wmv', 'aiff', 'alac', 'flac', 'm4a', 'mka', 'mp3', 'ogg', 'opus', 'wav', 'aac', 'ape', 'asf', 'f4a', 'f4b']
    + ['m4b', 'm4p', 'm4r', 'oga', 'ogx', 'spx', 'vorbis', 'wma', 'weba', 'jpg', 'png', 'webp']
    + ['mhtml', 'srt', 'vtt', 'ass', 'lrc', 'f4f', 'f4m', 'm3u8', 'smil', 'mpd', 'pdf', 'doc', 'docx', 'excl']
    + ['aac', 'abw', 'arc', 'avif', 'avi', 'azw', 'bin', 'bmp', 'bz', 'bz2', 'cda', 'csh']
    + ['css', 'csv', 'doc', 'docx', 'eot', 'epub', 'gz', 'gif', 'htm', 'html', 'ico']
    + ['ics', 'jar', 'jpeg', 'jpg', 'js', 'json', 'jsonld', 'mid', 'midi', 'mjs', 'mp3']
    + ['mp4', 'mpeg', 'mpkg', 'odp', 'ods', 'odt', 'oga', 'ogv', 'ogx', 'opus', 'otf']
    + ['png', 'pdf', 'php', 'ppt', 'pptx', 'rar', 'rtf', 'sh', 'svg', 'tar', 'tif']
    + ['tiff', 'ts', 'ttf', 'txt', 'vsd', 'wav', 'weba', 'webm', 'webp', 'woff', 'woff2']
    + ['xhtml', 'xls', 'xlsx', 'xml', 'xml', 'xul', 'zip', '3gp', '3g2', '7z']
)


def determine_ext(url, default_ext='unknown_file'):
    if url is None or '.' not in url:
        return default_ext
    guess = url.partition('?')[0].rpartition('.')[2]
    if re.match(r'^[A-Za-z0-9]+$', guess):
        return guess
    # Try extract ext from URLs like http://example.com/foo/bar.mp4/?download
    elif guess.rstrip('/') in KNOWN_EXTENSIONS:
        return guess.rstrip('/')
    else:
        return default_ext


def timeconvert(timestr):
    """Convert RFC 2822 defined time string into system timestamp"""
    timestamp = None
    timetuple = email.utils.parsedate_tz(timestr)
    if timetuple is not None:
        timestamp = email.utils.mktime_tz(timetuple)
    return timestamp


def float_or_none(v, scale=1, invscale=1, default=None):
    if v is None:
        return default
    try:
        return float(v) * invscale / scale
    except (ValueError, TypeError):
        return default


def format_decimal_suffix(num, fmt='%d%s', *, factor=1000):
    """Formats numbers with decimal sufixes like K, M, etc"""
    num, factor = float_or_none(num), float(factor)
    if num is None or num < 0:
        return None
    POSSIBLE_SUFFIXES = 'kMGTPEZY'
    exponent = 0 if num == 0 else min(int(math.log(num, factor)), len(POSSIBLE_SUFFIXES))
    suffix = ['', *POSSIBLE_SUFFIXES][exponent]
    if factor == 1024:
        suffix = {'k': 'Ki', '': ''}.get(suffix, f'{suffix}i')
    converted = num / (factor**exponent)
    return fmt % (converted, suffix)


def format_bytes(bytes_to_format):
    return format_decimal_suffix(bytes_to_format, '%.2f%sB', factor=1024) or 'N/A'


# needed for sanitizing filenames in restricted mode
ACCENT_CHARS = dict(
    zip(
        'ÂÃÄÀÁÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖŐØŒÙÚÛÜŰÝÞßàáâãäåæçèéêëìíîïðñòóôõöőøœùúûüűýþÿ',
        itertools.chain(
            'AAAAAA',
            ['AE'],
            'CEEEEIIIIDNOOOOOOO',
            ['OE'],
            'UUUUUY',
            ['TH', 'ss'],
            'aaaaaa',
            ['ae'],
            'ceeeeiiiionooooooo',
            ['oe'],
            'uuuuuy',
            ['th'],
            'y',
        ),
    )
)

NO_DEFAULT = object()


def is_path_like(f):
    return isinstance(f, (str, bytes, os.PathLike))


def str_or_none(v, default=None):
    return default if v is None else str(v)


def convert_to_aiohttp_cookie_jar(mozilla_cookie_jar: http.cookiejar.MozillaCookieJar):
    """
    Convert an http.cookiejar.MozillaCookieJar that uses a Netscape HTTP Cookie File to an aiohttp.cookiejar.CookieJar
    Tested with aiohttp v3.8.4
    """
    aiohttp_cookie_jar = CookieJar(unsafe=True)  # unsafe = Allow also cookies for IPs

    # pylint: disable=protected-access
    for cookie_domain, domain_cookies in mozilla_cookie_jar._cookies.items():
        for cookie_path, path_cookies in domain_cookies.items():
            for cookie_name, cookie in path_cookies.items():
                # cookie_name is cookie.name; cookie_path is cookie.path; cookie_domain is cookie.domain
                morsel = http.cookies.Morsel()
                morsel.update(
                    {
                        "expires": cookie.expires,
                        "path": cookie.path,
                        "comment": cookie.comment,
                        "domain": cookie.domain,
                        # "max-age"  : "Max-Age",
                        "secure": cookie.secure,
                        # "httponly": "HttpOnly",
                        "version": cookie.version,
                        # "samesite": "SameSite",
                    }
                )
                # pylint: disable=protected-access
                morsel.set(cookie.name, cookie.value, http.cookies._quote(cookie.value))
                aiohttp_cookie_jar._cookies[(cookie_domain, cookie_path)][cookie_name] = morsel

    return aiohttp_cookie_jar


class MoodleDLCookieJar(http.cookiejar.MozillaCookieJar):
    """
    Taken from yt-dlp: Last update 9. Sep. 2022
    See [1] for cookie file format.

    1. https://curl.haxx.se/docs/http-cookies.html
    """

    _HTTPONLY_PREFIX = '#HttpOnly_'
    _ENTRY_LEN = 7
    _HEADER = '''# Netscape HTTP Cookie File
# This file is generated by moodle-dl.  Do not edit.

'''
    _CookieFileEntry = collections.namedtuple(
        'CookieFileEntry', ('domain_name', 'include_subdomains', 'path', 'https_only', 'expires_at', 'name', 'value')
    )

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(None, *args, **kwargs)
        if is_path_like(filename):
            filename = os.fspath(filename)
        self.filename = filename

    @staticmethod
    def _true_or_false(cndn):
        return 'TRUE' if cndn else 'FALSE'

    @contextlib.contextmanager
    def open(self, file, *, write=False):
        if is_path_like(file):
            with open(file, 'w' if write else 'r', encoding='utf-8') as f:
                yield f
        else:
            if write:
                file.truncate(0)
            yield file

    def _really_save(self, f, ignore_discard=False, ignore_expires=False):
        now = time.time()
        for cookie in self:
            if not ignore_discard and cookie.discard or not ignore_expires and cookie.is_expired(now):
                continue
            name, value = cookie.name, cookie.value
            if value is None:
                # cookies.txt regards 'Set-Cookie: foo' as a cookie
                # with no name, whereas http.cookiejar regards it as a
                # cookie with no value.
                name, value = '', name
            f.write(
                '%s\n'
                % '\t'.join(
                    (
                        cookie.domain,
                        self._true_or_false(cookie.domain.startswith('.')),
                        cookie.path,
                        self._true_or_false(cookie.secure),
                        str_or_none(cookie.expires, default=''),
                        name,
                        value,
                    )
                )
            )

    def save(self, filename=None, *args, **kwargs):
        """
        Save cookies to a file.
        Code is taken from CPython 3.6
        https://github.com/python/cpython/blob/8d999cbf4adea053be6dbb612b9844635c4dfb8e/Lib/http/cookiejar.py#L2091-L2117
        """

        if filename is None:
            if self.filename is not None:
                filename = self.filename
            else:
                raise ValueError(http.cookiejar.MISSING_FILENAME_TEXT)

        # Store session cookies with `expires` set to 0 instead of an empty string
        for cookie in self:
            if cookie.expires is None:
                cookie.expires = 0

        with self.open(filename, write=True) as f:
            f.write(self._HEADER)
            self._really_save(f, *args, **kwargs)

    def load(self, filename=None, ignore_discard=False, ignore_expires=False):
        """Load cookies from a file."""
        if filename is None:
            if self.filename is not None:
                filename = self.filename
            else:
                raise ValueError(http.cookiejar.MISSING_FILENAME_TEXT)

        def prepare_line(line):
            if line.startswith(self._HTTPONLY_PREFIX):
                line = line[len(self._HTTPONLY_PREFIX) :]
            # comments and empty lines are fine
            if line.startswith('#') or not line.strip():
                return line
            cookie_list = line.split('\t')
            if len(cookie_list) != self._ENTRY_LEN:
                raise http.cookiejar.LoadError('invalid length %d' % len(cookie_list))
            cookie = self._CookieFileEntry(*cookie_list)
            if cookie.expires_at and not cookie.expires_at.isdigit():
                raise http.cookiejar.LoadError('invalid expires at %s' % cookie.expires_at)
            return line

        cf = io.StringIO()
        with self.open(filename) as input_file:
            for line in input_file:
                try:
                    cf.write(prepare_line(line))
                except http.cookiejar.LoadError as cookie_err:
                    if f'{line.strip()} '[0] in '[{"':
                        raise http.cookiejar.LoadError(
                            'Cookies file must be Netscape formatted, not JSON. See  '
                            'https://github.com/C0D3D3V/Moodle-DL/wiki/Use-cookies-when-downloading'
                        )
                    logging.info('WARNING: Skipping cookie file entry due to %s: %r', cookie_err, line)
                    continue
        cf.seek(0)
        self._really_load(cf, filename, ignore_discard, ignore_expires)
        # Session cookies are denoted by either `expires` field set to
        # an empty string or 0. MozillaCookieJar only recognizes the former
        # (see [1]). So we need force the latter to be recognized as session
        # cookies on our own.
        # Session cookies may be important for cookies-based authentication,
        # e.g. usually, when user does not check 'Remember me' check box while
        # logging in on a site, some important cookies are stored as session
        # cookies so that not recognizing them will result in failed login.
        # 1. https://bugs.python.org/issue17164
        for cookie in self:
            # Treat `expires=0` cookies as session cookies
            if cookie.expires == 0:
                cookie.expires = None
                cookie.discard = True


class Timer:
    '''
    Timing Context Manager
    Can be used for future speed comparisons, like this:

    with Timer() as t:
        Do.stuff()
    print(f'Do.stuff() took:\t {t.duration:.3f} \tseconds.')
    '''

    def __init__(self, nanoseconds=False):
        self.start = 0.0
        self.duration = 0.0
        self.nanoseconds = nanoseconds

    def __enter__(self):
        if self.nanoseconds:
            self.start = time.perf_counter_ns()
        else:
            self.start = time.time()
        return self

    def __exit__(self, *args):
        if self.nanoseconds:
            end = time.perf_counter_ns()
            self.duration = (end - self.start) * 10**-9  # 1 nano-sec = 10^-9 sec
        else:
            end = time.time()
            self.duration = end - self.start


PathParts = collections.namedtuple('PathParts', ('dir_name', 'file_name', 'file_extension'))


class PathTools:
    """A set of methods to create correct paths."""

    restricted_filenames = False

    @staticmethod
    def to_valid_name(name: str, is_file: bool, max_length: int = 200) -> str:
        """
        Filtering invalid characters in filenames and paths.

        @param name: The string that will go through the filtering
        @param is_file: If true, it is tried to keep the extension of the file name
        @param max_length: Most filesystems allow a max filename length of 255 chars,
                            we default use a shorter name to allow long extensions
        @return: The filtered string, that can be used as a filename.
        """

        if name is None:
            return None

        # Moodle saves the title of a section in HTML-Format,
        # so we need to unescape the string
        name = html.unescape(name)

        name = name.replace('\n', ' ')
        name = name.replace('\r', ' ')
        name = name.replace('\t', ' ')
        name = name.replace('\xad', '')
        while '  ' in name:
            name = name.replace('  ', ' ')
        name = PathTools.sanitize_filename(name, PathTools.restricted_filenames)
        name = name.strip('. ')
        name = name.strip()
        name = PathTools.truncate_filename(name, is_file, max_length)

        return name

    @staticmethod
    def truncate_filename(name: str, is_file: bool, max_length: int):
        if len(name) > max_length:
            if not is_file:
                name = PathTools.truncate_name(name, max_length)
            else:
                stem, ext = PathTools.get_file_stem_and_ext(name)
                ext_len = len(ext)
                if ext is None or ext_len == 0 or ext_len > 20:
                    # extensions longer then 20 characters are probably no extensions
                    name = PathTools.truncate_name(name, max_length)
                else:
                    stem = PathTools.truncate_name(stem, max_length - ext_len - 1)
                    name = f'{stem}.{ext}'
        return name

    @staticmethod
    def truncate_name(name: str, max_length: int):
        if PathTools.restricted_filenames:
            name = name[: max_length - 3] + '...'
        else:
            name = name[: max_length - 1] + '…'
        return name

    @staticmethod
    def remove_start(s, start):
        return s[len(start) :] if s is not None and s.startswith(start) else s

    @staticmethod
    def sanitize_filename(s, restricted=False, is_id=NO_DEFAULT):
        """Sanitizes a string so it could be used as part of a filename.
        @param restricted   Use a stricter subset of allowed characters
        @param is_id        Whether this is an ID that should be kept unchanged if possible.
                            If unset, yt-dlp's new sanitization rules are in effect
        """
        if s == '':
            return ''

        def replace_insane(char):
            if restricted and char in ACCENT_CHARS:
                return ACCENT_CHARS[char]
            elif not restricted and char == '\n':
                return '\0 '
            elif is_id is NO_DEFAULT and not restricted and char in '"*:<>?|/\\':
                # Replace with their full-width unicode counterparts
                return {'/': '\u29F8', '\\': '\u29f9'}.get(char, chr(ord(char) + 0xFEE0))
            elif char == '?' or ord(char) < 32 or ord(char) == 127:
                return ''
            elif char == '"':
                return '' if restricted else '\''
            elif char == ':':
                return '\0_\0-' if restricted else '\0 \0-'
            elif char in '\\/|*<>':
                return '\0_'
            if restricted and (char in '!&\'()[]{}$;`^,#' or char.isspace() or ord(char) > 127):
                return '\0_'
            return char

        if restricted and is_id is NO_DEFAULT:
            s = unicodedata.normalize('NFKC', s)
        s = re.sub(r'[0-9]+(?::[0-9]+)+', lambda m: m.group(0).replace(':', '_'), s)  # Handle timestamps
        result = ''.join(map(replace_insane, s))
        if is_id is NO_DEFAULT:
            result = re.sub(r'(\0.)(?:(?=\1)..)+', r'\1', result)  # Remove repeated substitute chars
            STRIP_RE = r'(?:\0.|[ _-])*'
            result = re.sub(f'^\0.{STRIP_RE}|{STRIP_RE}\0.$', '', result)  # Remove substitute chars from start/end
        result = result.replace('\0', '') or '_'

        if not is_id:
            while '__' in result:
                result = result.replace('__', '_')
            result = result.strip('_')
            # Common case of "Foreign band name - English song title"
            if restricted and result.startswith('-_'):
                result = result[2:]
            if result.startswith('-'):
                result = '_' + result[len('-') :]
            result = result.lstrip('.')
            if not result:
                result = '_'
        return result

    @staticmethod
    def sanitize_path(path: str):
        """
        @param path: A path to sanitize.
        @return: A path where every part was sanitized using to_valid_name.
        """
        drive_or_unc, _ = os.path.splitdrive(path)
        norm_path = os.path.normpath(PathTools.remove_start(path, drive_or_unc)).split(os.path.sep)
        if drive_or_unc:
            norm_path.pop(0)

        sanitized_path = [
            path_part if path_part in ['.', '..'] else PathTools.to_valid_name(path_part, is_file=False)
            for path_part in norm_path
        ]

        if drive_or_unc:
            sanitized_path.insert(0, drive_or_unc + os.path.sep)
        return os.path.join(*sanitized_path)

    @staticmethod
    def path_of_file_in_module(
        storage_path: str, course_fullname: str, file_section_name: str, file_module_name: str, file_path: str
    ):
        """
        @param storage_path: The path where all files should be stored.
        @param course_fullname: The name of the course where the file is
                                located.
        @param file_section_name: The name of the section where the file
                                  is located.
        @param file_module_name: The name of the module where the file
                                 is located.
        @param file_path: The additional path of a file (subdirectory).
        @return: A path where the file should be saved.
        """
        path = str(
            Path(storage_path)
            / PathTools.to_valid_name(course_fullname, is_file=False)
            / PathTools.to_valid_name(file_section_name, is_file=False)
            / PathTools.to_valid_name(file_module_name, is_file=False)
            / PathTools.sanitize_path(file_path).strip('/')
        )
        return path

    @staticmethod
    def path_of_file(storage_path: str, course_fullname: str, file_section_name: str, file_path: str):
        """
        @param storage_path: The path where all files should be stored.
        @param course_fullname: The name of the course where the file is
                                located.
        @param file_section_name: The name of the section where the file
                                  is located.
        @param file_path: The additional path of a file (subdirectory).
        @return: A path where the file should be saved.
        """
        path = str(
            Path(storage_path)
            / PathTools.to_valid_name(course_fullname, is_file=False)
            / PathTools.to_valid_name(file_section_name, is_file=False)
            / PathTools.sanitize_path(file_path).strip('/')
        )
        return path

    @staticmethod
    def flat_path_of_file(storage_path: str, course_fullname: str, file_path: str):
        """
        @param storage_path: The path where all files should be stored.
        @param course_fullname: The name of the course where the file is
                                located.
        @param file_path: The additional path of a file (subdirectory).
        @return: A path where the file should be saved.
        """
        path = str(
            Path(storage_path)
            / PathTools.to_valid_name(course_fullname, is_file=False)
            / PathTools.sanitize_path(file_path).strip('/')
        )
        return path

    @staticmethod
    def remove_file(file_path: str):
        if os.path.exists(file_path):
            os.unlink(file_path)

    @staticmethod
    def get_abs_path(path: str):
        return str(Path(path).resolve())

    @staticmethod
    def make_path(path: str, *filenames: str):
        result_path = Path(path)
        for filename in filenames:
            result_path = result_path / filename
        return str(result_path)

    @staticmethod
    def make_base_dir(path_to_file: str):
        Path(path_to_file).parent.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def make_dirs(path_to_dir: str):
        Path(path_to_dir).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def get_user_config_directory():
        """Returns a platform-specific root directory for user config settings."""
        # On Windows, prefer %LOCALAPPDATA%, then %APPDATA%, since we can expect the
        # AppData directories to be ACLed to be visible only to the user and admin
        # users (https://stackoverflow.com/a/7617601/1179226). If neither is set,
        # return None instead of falling back to something that may be world-readable.
        if os.name == "nt":
            appdata = os.getenv("LOCALAPPDATA")
            if appdata:
                return appdata
            appdata = os.getenv("APPDATA")
            if appdata:
                return appdata
            return None
        # On non-windows, use XDG_CONFIG_HOME if set, else default to ~/.config.
        xdg_config_home = os.getenv("XDG_CONFIG_HOME")
        if xdg_config_home:
            return xdg_config_home
        return os.path.join(os.path.expanduser("~"), ".config")

    @staticmethod
    def get_user_data_directory():
        """Returns a platform-specific root directory for user application data."""
        if os.name == "nt":
            appdata = os.getenv("LOCALAPPDATA")
            if appdata:
                return appdata
            appdata = os.getenv("APPDATA")
            if appdata:
                return appdata
            return None
        # On non-windows, use XDG_DATA_HOME if set, else default to ~/.config.
        xdg_config_home = os.getenv("XDG_DATA_HOME")
        if xdg_config_home:
            return xdg_config_home
        return os.path.join(os.path.expanduser("~"), ".local/share")

    @staticmethod
    def get_project_data_directory():
        """
        Returns an Path object to the project config directory
        """
        data_dir = Path(PathTools.get_user_data_directory()) / "moodle-dl"
        if not data_dir.is_dir():
            data_dir.mkdir(parents=True, exist_ok=True)
        return str(data_dir)

    @staticmethod
    def get_project_config_directory():
        """
        Returns an Path object to the project config directory
        """
        config_dir = Path(PathTools.get_user_config_directory()) / "moodle-dl"
        if not config_dir.is_dir():
            config_dir.mkdir(parents=True, exist_ok=True)
        return str(config_dir)

    @staticmethod
    def get_unused_filename(destination: str, filename: str, file_extension: str, start_clear=True):
        count = 0
        if start_clear:
            new_file_path = str(Path(destination) / f'{filename}.{file_extension}')
        else:
            new_file_path = str(Path(destination) / f'{filename}_{count:02d}.{file_extension}')
        while os.path.exists(new_file_path):
            count += 1
            new_file_path = str(Path(destination) / f'{filename}_{count:02d}.{file_extension}')

        return new_file_path

    @staticmethod
    def get_path_parts(file_path: str) -> PathParts:
        destination = os.path.dirname(file_path)
        filename, file_extension = os.path.splitext(os.path.basename(file_path))
        if file_extension.startswith('.'):
            file_extension = file_extension[1:]
        return PathParts(destination, filename, file_extension)

    @classmethod
    def get_unused_file_path(cls, file_path: str, start_clear=True):
        destination, filename, file_extension = cls.get_path_parts(file_path)
        return cls.get_unused_filename(destination, filename, file_extension, start_clear)

    @classmethod
    def touch_file(cls, file_path: str):
        open(file_path, 'a', encoding='utf-8').close()

    @staticmethod
    def get_file_exts(filename: str) -> (str, str):
        file_splits = filename.rsplit('.', 2)
        if len(file_splits) == 2:
            return None, file_splits[-1].lower()
        if len(file_splits) == 3:
            return file_splits[-2].lower(), file_splits[-1].lower()
        return None, None

    @staticmethod
    def get_file_ext(filename: str) -> str:
        file_splits = filename.rsplit('.', 1)
        if len(file_splits) == 2:
            return file_splits[-1].lower()
        return None

    @staticmethod
    def get_file_stem_and_ext(filename: str) -> (str, str):
        file_splits = filename.rsplit('.', 1)
        if len(file_splits) == 2:
            return file_splits[0], file_splits[1]
        return file_splits[0], None

    @staticmethod
    def get_cookies_path(storage_path: str) -> str:
        return str(Path(storage_path) / 'Cookies.txt')


class SslHelper:
    warned_about_certifi = False

    @classmethod
    def load_default_certs(cls, ssl_context: ssl.SSLContext):
        cert_loc = extract_zipped_paths(DEFAULT_CA_BUNDLE_PATH)

        if not cert_loc or not os.path.exists(cert_loc):
            if not cls.warned_about_certifi:
                Log.warning(f"Certifi could not find a suitable TLS CA certificate bundle, invalid path: {cert_loc}")
                cls.warned_about_certifi = True
            ssl_context.load_default_certs()
        else:
            if not os.path.isdir(cert_loc):
                ssl_context.load_verify_locations(cafile=cert_loc)
            else:
                ssl_context.load_verify_locations(capath=cert_loc)

    @classmethod
    @lru_cache(maxsize=4)
    def get_ssl_context(cls, skip_cert_verify: bool, allow_insecure_ssl: bool):
        if not skip_cert_verify:
            ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            cls.load_default_certs(ssl_context)
        else:
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.options |= ssl.OP_NO_SSLv2
            ssl_context.options |= ssl.OP_NO_SSLv3
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE
            try:
                ssl_context.options |= ssl.OP_NO_COMPRESSION
            except AttributeError as attr_err:
                Log.warning(
                    f"{attr_err!s}: The Python interpreter is compiled "
                    "against OpenSSL < 1.0.0. Ref: "
                    "https://docs.python.org/3/library/ssl.html"
                    "#ssl.OP_NO_COMPRESSION",
                )
            ssl_context.load_default_certs()

        if allow_insecure_ssl:
            # This allows connections to legacy insecure servers
            # https://www.openssl.org/docs/manmaster/man3/SSL_CTX_set_options.html#SECURE-RENEGOTIATION
            # Be warned the insecure renegotiation allows an attack, see:
            # https://nvd.nist.gov/vuln/detail/CVE-2009-3555
            ssl_context.options |= 0x4  # set ssl.OP_LEGACY_SERVER_CONNECT bit

        return ssl_context

    class CustomHttpAdapter(requests.adapters.HTTPAdapter):
        '''
        Transport adapter that allows us to use custom ssl_context.
        See https://stackoverflow.com/a/71646353 for more details.
        '''

        def __init__(self, ssl_context=None, **kwargs):
            self.ssl_context = ssl_context
            super().__init__(**kwargs)

        def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
            self.poolmanager = urllib3.poolmanager.PoolManager(
                num_pools=connections, maxsize=maxsize, block=block, ssl_context=self.ssl_context, **pool_kwargs
            )

    @classmethod
    def custom_requests_session(cls, skip_cert_verify: bool, allow_insecure_ssl: bool):
        """
        Return a new requests session with custom SSL context
        """
        session = requests.Session()
        ssl_context = cls.get_ssl_context(skip_cert_verify, allow_insecure_ssl)
        session.mount('https://', cls.CustomHttpAdapter(ssl_context))
        return session


class ProcessLock:
    """
    A very simple lock mechanism to prevent multiple downloader being started for the same Moodle.

    The functions are not resistant to high frequency calls.
    Raise conditions will occur!
    """

    class LockError(Exception):
        """An Exception which gets thrown if a Downloader is already running."""

        pass

    @staticmethod
    def lock(dir_path: str):
        """
        Test if a lock is already set in a directory, if not it creates the lock.
        """
        path = Path(dir_path) / 'running.lock'
        if Path(path).exists():
            raise ProcessLock.LockError(
                f'A downloader is already running. Delete {str(path)} if you think this is wrong.'
            )
        Path(path).touch()

    @staticmethod
    def unlock(dir_path: str):
        """Remove a lock in a directory."""
        path = Path(dir_path) / 'running.lock'
        try:
            Path(path).unlink()
        except OSError:
            pass


RESET_SEQ = '\033[0m'
COLOR_SEQ = '\033[1;%dm'

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(30, 38)


class Log:
    """
    Logs a given string to output with colors
    :param logString: the string that should be logged

    The string functions returns the strings that would be logged.

    The idea is to use this Log class only for the CLI configuration wizard,
    and for all other logging we use the normal python logging module
    """

    @staticmethod
    def info_str(logString: str):
        return COLOR_SEQ % WHITE + logString + RESET_SEQ

    @staticmethod
    def success_str(logString: str):
        return COLOR_SEQ % GREEN + logString + RESET_SEQ

    @staticmethod
    def green_str(logString: str):
        return COLOR_SEQ % GREEN + logString + RESET_SEQ

    @staticmethod
    def warning_str(logString: str):
        return COLOR_SEQ % YELLOW + logString + RESET_SEQ

    @staticmethod
    def yellow_str(logString: str):
        return COLOR_SEQ % YELLOW + logString + RESET_SEQ

    @staticmethod
    def error_str(logString: str):
        return COLOR_SEQ % RED + logString + RESET_SEQ

    @staticmethod
    def debug_str(logString: str):
        return COLOR_SEQ % CYAN + logString + RESET_SEQ

    @staticmethod
    def cyan_str(logString: str):
        return COLOR_SEQ % CYAN + logString + RESET_SEQ

    @staticmethod
    def blue_str(logString: str):
        return COLOR_SEQ % BLUE + logString + RESET_SEQ

    @staticmethod
    def magenta_str(logString: str):
        return COLOR_SEQ % MAGENTA + logString + RESET_SEQ

    @staticmethod
    def info(logString: str):
        print(Log.info_str(logString))

    @staticmethod
    def success(logString: str):
        print(Log.success_str(logString))

    @staticmethod
    def warning(logString: str):
        print(Log.warning_str(logString))

    @staticmethod
    def yellow(logString: str):
        print(Log.yellow_str(logString))

    @staticmethod
    def error(logString: str):
        print(Log.error_str(logString))

    @staticmethod
    def debug(logString: str):
        print(Log.debug_str(logString))

    @staticmethod
    def blue(logString: str):
        print(Log.blue_str(logString))

    @staticmethod
    def magenta(logString: str):
        print(Log.magenta_str(logString))

    @staticmethod
    def cyan(logString: str):
        print(Log.cyan_str(logString))


class Cutie:
    """
    Command-line User Tools for simpler Input

    Source: https://github.com/Kamik423/cutie/blob/master/cutie.py

    Authors: Hans and Kamik423
    License: MIT (https://github.com/Kamik423/cutie/blob/master/license.md)
    """

    class DefaultKeys:
        """List of default keybindings.
        Attributes:
            interrupt(List[str]): Keys that cause a keyboard interrupt.
            select(List[str]): Keys that trigger list element selection.
            select_all(List[str]): Keys that trigger selection of all list elements.
            confirm(List[str]): Keys that trigger list confirmation.
            delete(List[str]): Keys that trigger character deletion.
            down(List[str]): Keys that select the element below.
            up(List[str]): Keys that select the element above.
        """

        interrupt: List[str] = [readchar.key.CTRL_C, readchar.key.CTRL_D]
        select: List[str] = [readchar.key.SPACE]
        select_all: List[str] = [readchar.key.CTRL_A]
        confirm: List[str] = [readchar.key.ENTER, readchar.key.CR, readchar.key.LF]
        delete: List[str] = [readchar.key.BACKSPACE]
        down: List[str] = [readchar.key.DOWN, 'j']
        up: List[str] = [readchar.key.UP, 'k']
        tab: List[str] = ['\t']

    @staticmethod
    def get_number(
        prompt: str, min_value: Optional[float] = None, max_value: Optional[float] = None, allow_float: bool = True
    ) -> float:
        """Get a number from user input.
        If an invalid number is entered the user will be prompted again.
        Args:
            prompt (str): The prompt asking the user to input.
            min_value (float, optional): The [inclusive] minimum value.
            max_value (float, optional): The [inclusive] maximum value.
            allow_float (bool, optional): Allow floats or force integers.
        Returns:
            float: The number input by the user.
        """
        return_value: Optional[float] = None
        while return_value is None:
            input_value = input(prompt + ' ')
            try:
                return_value = float(input_value)
            except ValueError:
                print('Not a valid number.\033[K\033[1A\r\033[K', end='')
            if not allow_float and return_value is not None:
                if return_value != int(return_value):
                    print('Has to be an integer.\033[K\033[1A\r\033[K', end='')
                    return_value = None
            if min_value is not None and return_value is not None:
                if return_value < min_value:
                    print(f'Has to be at least {min_value}.\033[K\033[1A\r\033[K', end='')
                    return_value = None
            if max_value is not None and return_value is not None:
                if return_value > max_value:
                    print(f'Has to be at most {max_value}.\033[1A\r\033[K', end='')
                    return_value = None
            if return_value is not None:
                break
        print('\033[K', end='')
        if allow_float:
            return return_value
        return int(return_value)

    @staticmethod
    def secure_input(prompt: str) -> str:
        """Get secure input without showing it in the command line.
        Args:
            prompt (str): The prompt asking the user to input.
        Returns:
            str: The secure input.
        """
        return getpass.getpass(prompt + ' ')

    @staticmethod
    def select(
        options: List[str],
        caption_indices: Optional[List[int]] = None,
        deselected_prefix: str = '\033[1m[ ]\033[0m ',
        selected_prefix: str = '\033[1m[\033[32;1mx\033[0;1m]\033[0m ',
        caption_prefix: str = '',
        selected_index: int = 0,
        confirm_on_select: bool = True,
        reserved_lines: int = 3,
    ) -> int:
        """Select an option from a list.
        Args:
            options (List[str]): The options to select from.
            caption_indices (List[int], optional): Non-selectable indices.
            deselected_prefix (str, optional): Prefix for deselected option ([ ]).
            selected_prefix (str, optional): Prefix for selected option ([x]).
            caption_prefix (str, optional): Prefix for captions ().
            selected_index (int, optional): The index to be selected at first.
            confirm_on_select (bool, optional): Select keys also confirm.
            reserved_lines (int, optional): How many lines of the terminal are
                reserved and should not be used.
        Returns:
            int: The index that has been selected.
        """
        if caption_indices is None:
            caption_indices = []

        max_index = len(options)
        max_lines = len(options) + 1  # Last line is for empty line

        # Lines that were output in the previous interation
        lines_printed = 0
        # By how many entries is the list shifted
        shift = 0

        while True:
            print(f'\033[{lines_printed}A')

            console_lines = shutil.get_terminal_size().lines
            # Extra empty line for correct terminal behavior
            view_hight = max(0, min(console_lines - reserved_lines, max_lines))
            # View consists of
            # (top-indicator)
            # entries
            # (bottom-indicator)
            # empty line

            #  Darstellbaren Einträge =  view_hight - top-indicator
            if shift > (max_lines - 1) - (view_hight - 2):
                shift = (max_lines - 1) - (view_hight - 2)

            # Darstellbaren Einträge normal =  view_hight - top-indicator - bottom-indicator - empty line
            data_bottom = shift + (view_hight - 3)
            if shift == 0:
                # we do not need to print "x more lines above...", so we have one more entry line
                data_bottom += 1
            else:
                print(f'\033[K{shift} more lines above...')

            if data_bottom + 1 >= len(options):
                data_bottom = len(options)

            for i in range(shift, data_bottom):
                option = options[i]
                console_columns = shutil.get_terminal_size().columns - 5
                printable_option = option.expandtabs().replace('\n', ' ').replace('\r', ' ')
                if len(printable_option) > console_columns:
                    printable_option = printable_option[: (console_columns - 2)] + '..'

                if i not in caption_indices:
                    print(f'\033[K{selected_prefix if i == selected_index else deselected_prefix}{printable_option}')
                elif i in caption_indices:
                    print(f'\033[K{caption_prefix}{printable_option}')

            if data_bottom != len(options):
                print(f'{len(options) - data_bottom} more lines below...\033[K')

            lines_printed = view_hight

            keypress = readchar.readkey()
            if keypress in Cutie.DefaultKeys.up:
                new_index = selected_index
                while new_index > 0:
                    new_index -= 1
                    if new_index not in caption_indices:
                        selected_index = new_index
                        if selected_index < shift:
                            if shift == 2:
                                shift = 0
                            else:
                                shift = selected_index
                        break
            elif keypress in Cutie.DefaultKeys.down:
                new_index = selected_index
                while new_index < max_index - 1:
                    new_index += 1
                    if new_index not in caption_indices:
                        selected_index = new_index
                        if selected_index >= data_bottom and data_bottom != max_index:
                            shift = selected_index - (view_hight - 4)
                        break
            elif keypress in Cutie.DefaultKeys.confirm or confirm_on_select and keypress in Cutie.DefaultKeys.select:
                break
            elif keypress in Cutie.DefaultKeys.interrupt:
                raise KeyboardInterrupt
        return selected_index

    @staticmethod
    def select_multiple(
        options: List[str],
        caption_indices: Optional[List[int]] = None,
        deselected_unticked_prefix: str = '\033[1m( )\033[0m ',
        deselected_ticked_prefix: str = '\033[1m(\033[32mx\033[0;1m)\033[0m ',
        selected_unticked_prefix: str = '\033[32;1m{ }\033[0m ',
        selected_ticked_prefix: str = '\033[32;1m{x}\033[0m ',
        caption_prefix: str = '',
        ticked_indices: Optional[List[int]] = None,
        cursor_index: int = 0,
        minimal_count: int = 0,
        maximal_count: Optional[int] = None,
        hide_confirm: bool = False,
        deselected_confirm_label: str = '\033[1m(( confirm ))\033[0m',
        selected_confirm_label: str = '\033[1;32m{{ confirm }}\033[0m',
        reserved_lines: int = 3,
    ) -> List[int]:
        """Select multiple options from a list.
        Args:
            options (List[str]): The options to select from.
            caption_indices (List[int], optional): Non-selectable indices.
            deselected_unticked_prefix (str, optional): Prefix for lines that are
                not selected and not ticked (( )).
            deselected_ticked_prefix (str, optional): Prefix for lines that are
                not selected but ticked ((x)).
            selected_unticked_prefix (str, optional): Prefix for lines that are
                selected but not ticked ({ }).
            selected_ticked_prefix (str, optional): Prefix for lines that are
                selected and ticked ({x}).
            caption_prefix (str, optional): Prefix for captions ().
            ticked_indices (List[int], optional): Indices that are
                ticked initially.
            cursor_index (int, optional): The index the cursor starts at.
            minimal_count (int, optional): The minimal amount of lines
                that have to be ticked.
            maximal_count (int, optional): The maximal amount of lines
                that have to be ticked.
            hide_confirm (bool, optional): Hide the confirm button.
                This causes <ENTER> to confirm the entire selection and not just
                tick the line.
            deselected_confirm_label (str, optional): The confirm label
                if not selected ((( confirm ))).
            selected_confirm_label (str, optional): The confirm label
                if selected ({{ confirm }}).
            reserved_lines (int, optional): How many lines of the terminal are
                reserved and should not be used.
        Returns:
            List[int]: The indices that have been selected
        """
        if caption_indices is None:
            caption_indices = []
        if ticked_indices is None:
            ticked_indices = []
        max_index = len(options) - (1 if hide_confirm else 0)
        max_lines = len(options) + 2  # Last two line are for confirm / error / bottom-indicator + empty line
        error_message = ''

        # Lines that were output in the previous interation
        lines_printed = 0
        # By how many entries is the list shifted
        shift = 0

        while True:
            print(f'\033[{lines_printed}A')

            console_lines = shutil.get_terminal_size().lines
            # Extra empty line for correct terminal behavior
            view_hight = max(0, min(console_lines - reserved_lines, max_lines))
            # View consists of
            # (top-indicator)
            # entries
            # (bottom-indicator / error / confirm)
            # empty line

            #  Darstellbaren Einträge =  view_hight - bottom-indicator/confirm/error - top-indicator
            if shift > (max_lines - 1) - (view_hight - 2):
                shift = (max_lines - 1) - (view_hight - 2)

            # Darstellbaren Einträge normal =  view_hight - top-indicator - bottom-indicator/confirm/error - empty line
            data_bottom = shift + (view_hight - 3)
            if shift == 0:
                # we do not need to print "x more lines above...", so we have one more entry line
                data_bottom += 1
            else:
                print(f'\033[K{shift} more lines above...')

            if data_bottom > len(options):
                data_bottom = len(options)

            for i in range(shift, data_bottom):
                option = options[i]
                console_columns = shutil.get_terminal_size().columns - 5
                printable_option = option.expandtabs().replace('\n', ' ').replace('\r', ' ')
                if len(printable_option) > console_columns:
                    printable_option = printable_option[: (console_columns - 2)] + '..'

                prefix = ''
                if i in caption_indices:
                    prefix = caption_prefix
                elif i == cursor_index:
                    if i in ticked_indices:
                        prefix = selected_ticked_prefix
                    else:
                        prefix = selected_unticked_prefix
                else:
                    if i in ticked_indices:
                        prefix = deselected_ticked_prefix
                    else:
                        prefix = deselected_unticked_prefix
                print(f'\033[K{prefix}{printable_option}')

            if data_bottom == len(options):
                # we do not need to print "x more lines below...", instead we print the confirm label or an error
                if hide_confirm:
                    print(f'{error_message}\033[K')
                else:
                    if cursor_index == max_index:
                        print(f'{selected_confirm_label} {error_message}\033[K')
                    else:
                        print(f'{deselected_confirm_label} {error_message}\033[K')
            else:
                print(f'{len(options) - data_bottom} more lines below... {error_message}\033[K')

            lines_printed = view_hight

            error_message = ''
            keypress = readchar.readkey()
            if keypress in Cutie.DefaultKeys.up:
                new_index = cursor_index
                while new_index > 0:
                    new_index -= 1
                    if new_index not in caption_indices:
                        cursor_index = new_index
                        if cursor_index < shift:
                            if shift == 2:
                                shift = 0
                            else:
                                shift = cursor_index
                        break
            elif keypress in Cutie.DefaultKeys.down:
                new_index = cursor_index
                while new_index + 1 <= max_index:
                    new_index += 1
                    if new_index not in caption_indices:
                        cursor_index = new_index
                        if cursor_index >= data_bottom and data_bottom != len(options):
                            shift = cursor_index - (view_hight - 4)
                        break
            elif keypress in Cutie.DefaultKeys.select:
                if cursor_index in ticked_indices:
                    if len(ticked_indices) - 1 >= minimal_count:
                        ticked_indices.remove(cursor_index)
                elif maximal_count is not None:
                    if len(ticked_indices) + 1 <= maximal_count:
                        ticked_indices.append(cursor_index)
                else:
                    ticked_indices.append(cursor_index)
            elif keypress in Cutie.DefaultKeys.confirm:
                if minimal_count > len(ticked_indices):
                    error_message = f'Must select at least {minimal_count} options'
                elif maximal_count is not None and maximal_count < len(ticked_indices):
                    error_message = f'Must select at most {maximal_count} options'
                else:
                    break
            elif keypress in Cutie.DefaultKeys.select_all:
                for i in range(0, len(options)):
                    if i not in ticked_indices:
                        ticked_indices.append(i)
            elif keypress in Cutie.DefaultKeys.interrupt:
                raise KeyboardInterrupt
        print('\033[1A\033[K', end='', flush=True)
        return ticked_indices

    @staticmethod
    def prompt_yes_or_no(
        question: str,
        yes_text: str = 'Yes',
        no_text: str = 'No',
        has_to_match_case: bool = False,
        enter_empty_confirms: bool = True,
        default_is_yes: bool = False,
        deselected_prefix: str = '  ',
        selected_prefix: str = '\033[31m>\033[0m ',
        char_prompt: bool = True,
    ) -> Optional[bool]:
        """Prompt the user to input yes or no.
        Args:
            question (str): The prompt asking the user to input.
            yes_text (str, optional): The text corresponding to 'yes'.
            no_text (str, optional): The text corresponding to 'no'.
            has_to_match_case (bool, optional): Does the case have to match.
            enter_empty_confirms (bool, optional): Does enter on empty string work.
            default_is_yes (bool, optional): Is yes selected by default (no).
            deselected_prefix (str, optional): Prefix if something is deselected.
            selected_prefix (str, optional): Prefix if something is selected (> )
            char_prompt (bool, optional): Add a [Y/N] to the prompt.
        Returns:
            Optional[bool]: The bool what has been selected.
        """
        is_yes = default_is_yes
        is_selected = enter_empty_confirms
        current_message = ''
        if char_prompt:
            if default_is_yes:
                yn_prompt = f' ({yes_text[0]}/{no_text[0].lower()}) '
            else:
                yn_prompt = f' ({yes_text[0].lower()}/{no_text[0]}) '
        else:
            yn_prompt = ': '
        print()
        while True:
            yes = is_yes and is_selected
            no = not is_yes and is_selected
            print('\033[K' f'{selected_prefix if yes else deselected_prefix}{yes_text}')
            print('\033[K' f'{selected_prefix if no else deselected_prefix}{no_text}')
            print('\033[3A\r\033[K' f'{question}{yn_prompt}{current_message}', end='', flush=True)
            keypress = readchar.readkey()
            if keypress in Cutie.DefaultKeys.down or keypress in Cutie.DefaultKeys.up:
                is_yes = not is_yes
                is_selected = True
                current_message = yes_text if is_yes else no_text
            elif keypress in Cutie.DefaultKeys.interrupt:
                raise KeyboardInterrupt
            elif keypress in Cutie.DefaultKeys.confirm:
                if is_selected:
                    break
            elif keypress is not None:
                if keypress in Cutie.DefaultKeys.tab:
                    pass
                elif keypress in Cutie.DefaultKeys.delete:
                    if current_message:
                        current_message = current_message[:-1]
                else:
                    current_message += keypress

                match_yes = yes_text
                match_no = no_text
                match_text = current_message
                if not has_to_match_case:
                    match_yes = match_yes.upper()
                    match_no = match_no.upper()
                    match_text = match_text.upper()
                if match_no.startswith(match_text):
                    is_selected = True
                    is_yes = False
                elif match_yes.startswith(match_text):
                    is_selected = True
                    is_yes = True
                else:
                    is_selected = False

                if keypress in Cutie.DefaultKeys.tab and is_selected:
                    current_message = yes_text if is_yes else no_text
            print()
        print('\033[K\n\033[K\n\033[K\n\033[3A')
        return is_selected and is_yes



from typing import List

from yt_dlp.YoutubeDL import YoutubeDL
from yt_dlp.extractor.common import InfoExtractor

from moodle_dl.downloader.extractors.googledrive import GoogleDriveIE  # noqa: F401
from moodle_dl.downloader.extractors.helixmedia_lti import HelixmediaLtiIE  # noqa: F401
from moodle_dl.downloader.extractors.kalvidres_lti import KalvidresLtiIE  # noqa: F401
from moodle_dl.downloader.extractors.opencast_lti import OpencastLtiIE  # noqa: F401
from moodle_dl.downloader.extractors.owncloud import OwncloudIE  # noqa: F401

ALL_ADDITIONAL_EXTRACTORS = [Class for name, Class in globals().items() if name.endswith('IE')]


def add_additional_extractors(ydl: YoutubeDL) -> List[InfoExtractor]:
    moodle_dl_ies = {}
    moodle_dl_ies_instances = {}
    for extractor_class in ALL_ADDITIONAL_EXTRACTORS:
        extractor = extractor_class(ydl)
        ie_key = extractor.ie_key()
        moodle_dl_ies[ie_key] = extractor
        moodle_dl_ies_instances[ie_key] = extractor

    # We access protected member variables of the yt-dlp to add the extractors afterwards.
    # TODO: Use the new possibilities yt-dlp offers to add the extractors to yt-dlp.
    # pylint: disable=protected-access
    moodle_dl_ies.update(ydl._ies)
    moodle_dl_ies_instances.update(ydl._ies_instances)
    ydl._ies = moodle_dl_ies
    ydl._ies_instances = moodle_dl_ies_instances



from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.types import Course, File


class PageMod(MoodleMod):
    MOD_NAME = 'page'
    MOD_PLURAL_NAME = 'pages'
    MOD_MIN_VERSION = 2017051500  # 3.3

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        # TODO: Add download condition
        return True

    async def real_fetch_mod_entries(self, courses: List[Course]) -> Dict[int, Dict[int, Dict]]:
        pages = (
            await self.client.async_post(
                'mod_page_get_pages_by_courses', self.get_data_for_mod_entries_endpoint(courses)
            )
        ).get('pages', [])

        result = {}
        for page in pages:
            course_id = page.get('course', 0)
            page_name = page.get('name', 'unnamed page')
            page_content = page.get('content', '')

            page_files = page.get('introfiles', [])
            page_files += page.get('contentfiles', [])
            self.set_files_types_if_empty(page_files, 'page_file')

            page_intro = page.get('intro', '')
            if page_intro != '':
                page_files.append(
                    {
                        'filename': 'Page intro',
                        'filepath': '/',
                        'description': page_intro,
                        'type': 'description',
                    }
                )

            if page_content != '':
                page_files.append(
                    {
                        'filename': page_name,
                        'filepath': '/',
                        'html': page_content,
                        'filter_urls_during_search_containing': ['/mod_page/content/'],
                        'no_hash': True,
                        'type': 'html',
                        'timemodified': page.get('timemodified', 0),
                        'filesize': len(page_content),
                    }
                )

            self.add_module(
                result,
                course_id,
                page.get('coursemodule', 0),
                {
                    'id': page.get('id', 0),
                    'name': page_name,
                    'files': page_files,
                },
            )

        return result



from abc import ABCMeta, abstractmethod
from typing import List

from moodle_dl.config import ConfigHelper
from moodle_dl.downloader.task import Task
from moodle_dl.types import Course


class NotificationService(metaclass=ABCMeta):
    "Common class for a notification service"

    def __init__(self, config: ConfigHelper):
        self.config = config

    @abstractmethod
    def notify_about_changes_in_moodle(self, changes: List[Course]) -> None:
        """
        Sends out a Notification to inform about detected changes for the
        Moodle-Account. The caller shouldn't care about if the sending was
        successful.
        @param changes: The detected changes per course.
        """
        pass

    @abstractmethod
    def notify_about_error(self, error_description: str) -> None:
        """
        Sends out a Notification to inform about an error encountered during
        the execution of the program.
        The caller shouldn't care about if the sending was successful.
        @param error_description: The error text.
        """
        pass

    @abstractmethod
    def notify_about_failed_downloads(self, failed_downloads: List[Task]) -> None:
        """
        Sends out a Notification to inform about failed downloads encountered during
        the execution of the program.
        The caller shouldn't care about if the sending was successful.
        @param failed_downloads: A list of failed Tasks.
        """
        pass



import argparse
import asyncio
import logging
import os
import sys
import traceback

from logging.handlers import RotatingFileHandler
from shutil import which

import colorlog
import requests  # noqa: F401 pylint: disable=unused-import
import sentry_sdk
import urllib3

try:
    # In unix readline needs to be loaded so that arrow keys work in input
    import readline  # pylint: disable=unused-import # noqa: F401
except ImportError:
    pass

from colorama import just_fix_windows_console

from moodle_dl.cli import ConfigWizard, DatabaseManager, MoodleWizard, NotificationsWizard, init_config
from moodle_dl.config import ConfigHelper
from moodle_dl.database import StateRecorder
from moodle_dl.downloader.download_service import DownloadService
from moodle_dl.downloader.fake_download_service import FakeDownloadService
from moodle_dl.moodle.moodle_service import MoodleService
from moodle_dl.notifications import get_all_notify_services
from moodle_dl.types import MoodleDlOpts
from moodle_dl.utils import ProcessLock, check_debug, PathTools as PT
from moodle_dl.version import __version__


class ReRaiseOnError(logging.StreamHandler):
    "A logging-handler class which allows the exception-catcher of i.e. PyCharm to intervene"

    def emit(self, record):
        if hasattr(record, 'exception'):
            raise record.exception


def choose_task(config: ConfigHelper, opts: MoodleDlOpts):
    if opts.add_all_visible_courses:
        ConfigWizard(config, opts).interactively_add_all_visible_courses()
    elif opts.change_notification_mail:
        NotificationsWizard(config, opts).interactively_configure_mail()
    elif opts.change_notification_telegram:
        NotificationsWizard(config, opts).interactively_configure_telegram()
    elif opts.change_notification_xmpp:
        NotificationsWizard(config, opts).interactively_configure_xmpp()
    elif opts.config:
        ConfigWizard(config, opts).interactively_acquire_config()
    elif opts.delete_old_files:
        DatabaseManager(config, opts).delete_old_files()
    elif opts.manage_database:
        DatabaseManager(config, opts).interactively_manage_database()
    elif opts.new_token:
        MoodleWizard(config, opts).interactively_acquire_token(use_stored_url=True)
    else:
        run_main(config, opts)


def connect_sentry(config: ConfigHelper) -> bool:
    "Return True if connected"
    try:
        sentry_dsn = config.get_property('sentry_dsn')
        if sentry_dsn:
            sentry_sdk.init(sentry_dsn)
            return True
    except (ValueError, sentry_sdk.utils.BadDsn, sentry_sdk.utils.ServerlessTimeoutWarning):
        pass
    return False


def run_main(config: ConfigHelper, opts: MoodleDlOpts):
    sentry_connected = connect_sentry(config)
    notify_services = get_all_notify_services(config)

    # TODO: Change this
    PT.restricted_filenames = config.get_restricted_filenames()

    try:
        moodle = MoodleService(config, opts)

        logging.debug('Checking for changes for the configured Moodle-Account....')
        database = StateRecorder(opts)
        changed_courses = asyncio.run(moodle.fetch_state(database))

        if opts.log_responses:
            logging.info("All JSON-responses from Moodle have been written to the responses.log file.")
            return

        logging.debug('Start downloading changed files...')

        if opts.without_downloading_files:
            downloader = FakeDownloadService(changed_courses, config, opts, database)
        else:
            downloader = DownloadService(changed_courses, config, opts, database)
        downloader.run()
        failed_downloads = downloader.get_failed_tasks()

        changed_courses_to_notify = database.changes_to_notify()

        if len(changed_courses_to_notify) > 0:
            for service in notify_services:
                service.notify_about_changes_in_moodle(changed_courses_to_notify)

            database.notified(changed_courses_to_notify)

        else:
            logging.info('No changes found for the configured Moodle-Account.')

        if len(failed_downloads) > 0:
            for service in notify_services:
                service.notify_about_failed_downloads(failed_downloads)

    except BaseException as base_err:
        if sentry_connected:
            sentry_sdk.capture_exception(base_err)

        short_error = str(base_err)
        if not short_error or short_error.isspace():
            short_error = traceback.format_exc(limit=1)

        for service in notify_services:
            service.notify_about_error(short_error)

        raise base_err


def setup_logger(opts: MoodleDlOpts):
    file_log_handler = RotatingFileHandler(
        PT.make_path(opts.path, 'MoodleDL.log'),
        mode='a',
        maxBytes=1 * 1024 * 1024,
        backupCount=2,
        encoding='utf-8',
        delay=0,
    )
    file_log_handler.setFormatter(
        logging.Formatter('%(asctime)s  %(levelname)s  {%(module)s}  %(message)s', '%Y-%m-%d %H:%M:%S')
    )
    stdout_log_handler = colorlog.StreamHandler()
    if sys.stdout.isatty() and not opts.verbose:
        stdout_log_handler.setFormatter(colorlog.ColoredFormatter('%(log_color)s%(asctime)s %(message)s', '%H:%M:%S'))
    else:
        stdout_log_handler.setFormatter(
            colorlog.ColoredFormatter(
                '%(log_color)s%(asctime)s  %(levelname)s  {%(module)s}  %(message)s', '%Y-%m-%d %H:%M:%S'
            )
        )

    app_log = logging.getLogger()
    if opts.quiet:
        file_log_handler.setLevel(logging.ERROR)
        app_log.setLevel(logging.ERROR)
        stdout_log_handler.setLevel(logging.ERROR)
    elif opts.verbose:
        file_log_handler.setLevel(logging.DEBUG)
        app_log.setLevel(logging.DEBUG)
        stdout_log_handler.setLevel(logging.DEBUG)
    else:
        file_log_handler.setLevel(logging.INFO)
        app_log.setLevel(logging.INFO)
        stdout_log_handler.setLevel(logging.INFO)

    app_log.addHandler(stdout_log_handler)
    if opts.log_to_file:
        app_log.addHandler(file_log_handler)

    if opts.verbose:
        logging.debug('moodle-dl version: %s', __version__)
        logging.debug('python version: %s', ".".join(map(str, sys.version_info[:3])))
        ffmpeg_available = which('ffmpeg') is not None
        logging.debug('Is ffmpeg available: %s', ffmpeg_available)

    if check_debug():
        logging.info('Debug-Mode detected. Errors will be re-risen.')
        app_log.addHandler(ReRaiseOnError())

    if not opts.verbose:
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger('asyncio').setLevel(logging.WARNING)
        urllib3.disable_warnings()


def _dir_path(path):
    if os.path.isdir(path):
        return path
    raise argparse.ArgumentTypeError(f'"{str(path)}" is not a valid path. Make sure the directory exists.')


def win_max_path_length_workaround(path):
    # Working around MAX_PATH limitation on Windows (see
    # http://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx)
    if os.name == 'nt':
        abs_file_path = PT.get_abs_path(path)
        path = '\\\\?\\' + abs_file_path
        logging.debug("Using windows max path length workaround")
    else:
        logging.info("You are not on Windows, you don't need to use this workaround")
    return path


def get_parser():
    """
    Creates a new argument parser.
    """
    parser = argparse.ArgumentParser(
        description=('Moodle-DL helps you download all the course files from your Moodle account.')
    )
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '-i',
        '--init',
        dest='init',
        default=False,
        action='store_true',
        help=(
            'Create an initial configuration. A CLI configuration wizard will lead you through'
            + ' the initial configuration.'
        ),
    )

    group.add_argument(
        '-c',
        '--config',
        dest='config',
        default=False,
        action='store_true',
        help=(
            'Start the configuration utility. It allows you to make almost all available moodle-dl settings'
            + ' conveniently via the CLI configuration wizard.'
        ),
    )

    group.add_argument(
        '-nt',
        '--new-token',
        dest='new_token',
        default=False,
        action='store_true',
        help=('Obtain a new login token. Use it if the saved token gets rejected by your Moodle.'),
    )

    group.add_argument(
        '-cm',
        '--change-notification-mail',
        dest='change_notification_mail',
        default=False,
        action='store_true',
        help=('Activate / deactivate / change the settings for receiving notifications via e-mail.'),
    )

    group.add_argument(
        '-ct',
        '--change-notification-telegram',
        dest='change_notification_telegram',
        default=False,
        action='store_true',
        help=('Activate / deactivate / change the settings for receiving notifications via Telegram.'),
    )

    group.add_argument(
        '-cx',
        '--change-notification-xmpp',
        dest='change_notification_xmpp',
        default=False,
        action='store_true',
        help=('Activate / deactivate / change the settings for receiving notifications via XMPP.'),
    )

    group.add_argument(
        '-md',
        '--manage-database',
        dest='manage_database',
        default=False,
        action='store_true',
        help=(
            'Manage the offline database. It allows you to delete entries from the database'
            + ' that are no longer available locally so that they can be downloaded again.'
        ),
    )

    group.add_argument(
        '-dof',
        '--delete-old-files',
        dest='delete_old_files',
        default=False,
        action='store_true',
        help=(
            'Delete old copies of files. It allows you to delete entries from the database'
            + ' and from local file system.'
        ),
    )

    group.add_argument(
        '--log-responses',
        dest='log_responses',
        default=False,
        action='store_true',
        help=(
            'Generate a responses.log file in which all JSON responses from your Moodle are logged'
            + ' along with the requested URLs.'
        ),
    )

    group.add_argument(
        '--add-all-visible-courses',
        dest='add_all_visible_courses',
        default=False,
        action='store_true',
        help='Add all courses visible to the user to the configuration file.',
    )

    group.add_argument(
        '--version',
        action='version',
        version='moodle-dl ' + __version__,
        help='Print program version and exit',
    )

    parser.add_argument(
        '-sso',
        '--sso',
        dest='sso',
        default=False,
        action='store_true',
        help=(
            'Use SSO login instead of normal login. This flag can be used together with --init and -nt.'
            + ' You will be guided through the Single Sign On (SSO) login process'
            + ' during initialization or new token retrieval.'
        ),
    )

    parser.add_argument(
        '-u',
        '--username',
        dest='username',
        default=None,
        type=str,
        help=('Specify username to skip the query when creating a new token.'),
    )

    parser.add_argument(
        '-pw',
        '--password',
        dest='password',
        default=None,
        type=str,
        help=('Specify password to skip the query when creating a new token.'),
    )

    parser.add_argument(
        '-tk',
        '--token',
        dest='token',
        default=None,
        type=str,
        help=('Specify token to skip the interactive login procedure.'),
    )
    parser.add_argument(
        '-p',
        '--path',
        dest='path',
        default='.',
        type=_dir_path,
        help=(
            'Sets the location of the configuration, logs and downloaded files. PATH must be an'
            + ' existing directory in which you have read and write access. (default: current working directory)'
        ),
    )

    parser.add_argument(
        '-mpac',
        '--max-parallel-api-calls',
        dest='max_parallel_api_calls',
        default=10,
        type=int,
        help=('Sets the number of max parallel Moodle Mobile API calls. (default: %(default)s)'),
    )

    parser.add_argument(
        '-mpd',
        '--max-parallel-downloads',
        dest='max_parallel_downloads',
        default=5,
        type=int,
        help=('Sets the number of max parallel downloads. (default: %(default)s)'),
    )

    parser.add_argument(
        '-mpyd',
        '--max-parallel-yt-dlp',
        dest='max_parallel_yt_dlp',
        default=5,
        type=int,
        help=('Sets the number of max parallel downloads using yt-dlp. (default: %(default)s)'),
    )

    parser.add_argument(
        '-dcs',
        '--download-chunk-size',
        dest='download_chunk_size',
        default=102400,
        type=int,
        help=('Sets the chunk size in bytes used when downloading files. (default: %(default)s)'),
    )

    parser.add_argument(
        '-iye',
        '--ignore-ytdl-errors',
        dest='ignore_ytdl_errors',
        default=False,
        action='store_true',
        help=(
            'Ignore errors that occur when downloading with the help of yt-dlp.'
            + ' Thus, no further attempt will be made to download the file using yt-dlp.'
            + ' By default, yt-dlp errors are critical, so the download of the corresponding file'
            + ' will be aborted and when you run moodle-dl again, the download will be repeated.'
        ),
    )

    parser.add_argument(
        '-wdf',
        '--without-downloading-files',
        dest='without_downloading_files',
        default=False,
        action='store_true',
        help=(
            'Do not download any file. This allows the local database to be updated'
            + ' without having to download all files.'
        ),
    )

    parser.add_argument(
        '--mpl',
        '--max-path-length-workaround',
        dest='max_path_length_workaround',
        default=False,
        action='store_true',
        help=(
            'Make all paths absolute in order to workaround the max_path limitation on Windows.'
            + ' To use relative paths on Windows you should disable the max_path limitation see:'
            + ' https://docs.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation'
        ),
    )

    parser.add_argument(
        '-ais',
        '--allow-insecure-ssl',
        dest='allow_insecure_ssl',
        default=False,
        action='store_true',
        help='Allow connections to unpatched servers. Use this option if your server uses a very old SSL version.',
    )
    parser.add_argument(
        '-scv',
        '--skip-cert-verify',
        dest='skip_cert_verify',
        default=False,
        action='store_true',
        help='Don\'t verify TLS certificates. This option should only be used in non production environments.',
    )

    parser.add_argument(
        '-v',
        '--verbose',
        dest='verbose',
        default=False,
        action='store_true',
        help='Print various debugging information',
    )

    parser.add_argument(
        '-q',
        '--quiet',
        dest='quiet',
        default=False,
        action='store_true',
        help='Sets the log level to error',
    )

    parser.add_argument(
        '-ltf',
        '--log-to-file',
        dest='log_to_file',
        default=False,
        action='store_true',
        help='Log all output additionally to a log file called MoodleDL.log',
    )

    return parser


def pre_process_opts(opts: MoodleDlOpts):
    if opts.max_path_length_workaround:
        opts.path = win_max_path_length_workaround(opts.path)

    # Max 32 yt-dlp threads
    opts.max_parallel_yt_dlp = min(opts.max_parallel_downloads, min(32, opts.max_parallel_yt_dlp))


# --- called at the program invocation: -------------------------------------
def main(args=None):
    """The main routine."""
    just_fix_windows_console()
    parser = get_parser()
    opts = MoodleDlOpts(**vars(parser.parse_args(args)))
    setup_logger(opts)
    pre_process_opts(opts)

    config = ConfigHelper(opts.path)
    if opts.init:
        init_config(config, opts)
        sys.exit(0)
    else:
        try:
            config.load()
        except ConfigHelper.NoConfigError as err_config:
            logging.error('Error: %s', err_config)
            logging.warning('You can create a configuration with the --init option')
            sys.exit(-1)

    try:
        if not check_debug():
            ProcessLock.lock(opts.path)

        choose_task(config, opts)

        logging.info('All done. Exiting..')
        ProcessLock.unlock(opts.path)
    except BaseException as base_err:  # pylint: disable=broad-except
        if not isinstance(base_err, ProcessLock.LockError):
            ProcessLock.unlock(opts.path)

        if opts.verbose or check_debug():
            logging.error(traceback.format_exc(), extra={'exception': base_err})
        else:
            logging.error('Exception: %s', base_err)

        logging.debug('Exception-Handling completed. Exiting...')

        sys.exit(1)



from getpass import getpass

from aioxmpp.errors import StanzaError, UserError

from moodle_dl.config import ConfigHelper
from moodle_dl.notifications.mail.mail_formater import create_full_welcome_mail
from moodle_dl.notifications.mail.mail_shooter import MailShooter
from moodle_dl.notifications.telegram.telegram_shooter import TelegramShooter, RequestRejectedError
from moodle_dl.notifications.xmpp.xmpp_shooter import XmppShooter
from moodle_dl.types import MoodleDlOpts
from moodle_dl.utils import Cutie


class NotificationsWizard:
    def __init__(self, config: ConfigHelper, opts: MoodleDlOpts):
        self.config = config
        self.opts = opts

    def interactively_configure_all_services(self) -> None:
        services = [
            getattr(self, func)
            for func in dir(self)
            if callable(getattr(self, func))
            and func.startswith("interactively_configure_")
            and func != "interactively_configure_all_services"
        ]
        for service in services:
            service()

    def interactively_configure_mail(self) -> None:
        "Guides the user through the configuration of the mail notification."

        do_mail = Cutie.prompt_yes_or_no('Do you want to activate Notifications via mail?')

        if not do_mail:
            self.config.remove_property('mail')
        else:
            print('[The following Inputs are not validated!]')

            config_valid = False
            while not config_valid:
                sender = input('E-Mail-Address of the Sender:   ')
                server_host = input('Host of the SMTP-Server:   ')
                server_port = input('Port of the SMTP-Server [STARTTLS, default 587]:   ')
                if server_port == '':
                    print('Using default port 587!')
                    server_port = '587'
                username = input('Username for the SMTP-Server:   ')
                password = getpass('Password for the SMTP-Server [no output]:   ')
                target = input('E-Mail-Address of the Target:   ')

                print('Testing Mail-Config...')
                welcome_content = create_full_welcome_mail()
                mail_shooter = MailShooter(sender, server_host, int(server_port), username, password)
                try:
                    mail_shooter.send(target, 'Hey!', welcome_content[0], welcome_content[1])
                except OSError as e:
                    print(f'Error while sending the test mail: {str(e)}')
                    continue
                else:
                    input(
                        'Please check if you received the Welcome-Mail.'
                        + ' If yes, confirm with Return.\nIf not, exit'
                        + ' this program ([CTRL]+[C]) and try again later.'
                    )
                    config_valid = True

                raw_send_error_msg = ''
                while raw_send_error_msg not in ['y', 'n']:
                    raw_send_error_msg = input('Do you want to also get error reports sent by mail? [y/n]   ')
                do_send_error_msg = raw_send_error_msg == 'y'

                mail_cfg = {
                    'sender': sender,
                    'server_host': server_host,
                    'server_port': server_port,
                    'username': username,
                    'password': password,
                    'target': target,
                    'send_error_msg': do_send_error_msg,
                }

                self.config.set_property('mail', mail_cfg)

    def interactively_configure_telegram(self) -> None:
        "Guides the user through the configuration of the telegram notification."

        do_telegram = Cutie.prompt_yes_or_no('Do you want to activate Notifications via Telegram?')

        if not do_telegram:
            self.config.remove_property('telegram')
        else:
            print('[The following Inputs are not validated!]')
            print(
                'Open the following link for help in setting up telegram notifications:'
                + ' https://github.com/C0D3D3V/Moodle-DL/wiki/Telegram-Notification'
            )
            config_valid = False
            while not config_valid:
                telegram_token = input('Telegram Token:    ')
                telegram_chatID = input('Telegram Chat ID:   ')

                print('Testing Telegram-Config...')

                try:
                    telegram_shooter = TelegramShooter(telegram_token, telegram_chatID)
                    telegram_shooter.send('This is a test message from moodle-dl!')
                except (ConnectionError, RuntimeError, RequestRejectedError) as e:
                    print(f'Error while sending the test message: {str(e)}')
                    continue

                else:
                    input(
                        'Please check if you received the Testmessage.'
                        + ' If yes, confirm with Return.\nIf not, exit'
                        + ' this program ([CTRL]+[C]) and try again later.'
                    )
                    config_valid = True

                raw_send_error_msg = ''
                while raw_send_error_msg not in ['y', 'n']:
                    raw_send_error_msg = input('Do you want to also get error reports sent in telegram? [y/n]   ')

                do_send_error_msg = raw_send_error_msg == 'y'

                telegram_cfg = {
                    'token': telegram_token,
                    'chat_id': telegram_chatID,
                    'send_error_msg': do_send_error_msg,
                }

                self.config.set_property('telegram', telegram_cfg)

    def interactively_configure_xmpp(self) -> None:
        "Guides the user through the configuration of the xmpp notification."

        do_xmpp = Cutie.prompt_yes_or_no('Do you want to activate Notifications via XMPP?')

        if not do_xmpp:
            self.config.remove_property('xmpp')
        else:
            print('[The following Inputs are not validated!]')
            config_valid = False
            while not config_valid:
                sender = input('JID of the Sender:   ')
                password = getpass('Password for the Sender [no output]:   ')
                target = input('JID of the Target:   ')
                print('Testing XMPP-Config...')

                try:
                    xmpp_shooter = XmppShooter(sender, password, target)
                    xmpp_shooter.send('This is a test message from moodle-dl!')
                except (
                    ConnectionError,
                    StanzaError,
                    UserError,
                    OSError,
                    RuntimeError,
                ) as e:
                    print(f'Error while sending the test message: {str(e)}')
                    continue

                else:
                    input(
                        'Please check if you received the Testmessage.'
                        + ' If yes, confirm with Return.\nIf not, exit'
                        + ' this program ([CTRL]+[C]) and try again later.'
                    )
                    config_valid = True

                raw_send_error_msg = ''
                while raw_send_error_msg not in ['y', 'n']:
                    raw_send_error_msg = input('Do you want to also get error reports sent in xmpp? [y/n]   ')

                do_send_error_msg = raw_send_error_msg == 'y'

                xmpp_cfg = {
                    'sender': sender,
                    'password': password,
                    'target': target,
                    'send_error_msg': do_send_error_msg,
                }

                self.config.set_property('xmpp', xmpp_cfg)



from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods import MoodleMod
from moodle_dl.types import Course, File


class AssignMod(MoodleMod):
    MOD_NAME = 'assign'
    MOD_PLURAL_NAME = 'assignments'
    MOD_MIN_VERSION = 2012120300  # 2.4

    @classmethod
    def download_condition(cls, config: ConfigHelper, file: File) -> bool:
        # TODO: Add condition for assignments not only submissions
        return config.get_download_submissions() or (not (file.module_modname.endswith(cls.MOD_NAME) and file.deleted))

    async def real_fetch_mod_entries(self, courses: List[Course]) -> Dict[int, Dict[int, Dict]]:
        assign_courses = (
            await self.client.async_post('mod_assign_get_assignments', self.get_data_for_mod_entries_endpoint(courses))
        ).get('courses', [])

        result = {}
        for assign_course in assign_courses:
            course_id = assign_course.get('id', 0)
            result[course_id] = self.extract_assign_modules(assign_course.get('assignments', []))

        await self.add_submissions(result)
        return result

    def extract_assign_modules(self, assignments: List[Dict]) -> Dict[int, Dict]:
        result = {}
        for assign in assignments:
            assign_files = assign.get('introfiles', [])
            assign_files += assign.get('introattachments', [])

            self.set_files_types_if_empty(assign_files, 'assign_file')

            assign_intro = assign.get('intro', '')
            if assign_intro != '':
                assign_files.append(
                    {
                        'filename': 'Assignment intro',
                        'filepath': '/',
                        'description': assign_intro,
                        'type': 'description',
                    }
                )

            result[assign.get('cmid', 0)] = {
                'id': assign.get('id', 0),
                'name': assign.get('name', ''),
                'timemodified': assign.get('timemodified', 0),
                'files': assign_files,
            }
        return result

    async def add_submissions(self, assignments: Dict[int, Dict[int, Dict]]):
        """
        Fetches for the assignments list additionally the submissions
        @param assignments: Dictionary of all assignments, indexed by courses, then module id
        """
        if not self.config.get_download_submissions():
            return

        if self.version < 2016052300:  # 3.1
            return

        await self.run_async_load_function_on_mod_entries(assignments, self.load_submissions)

    async def load_submissions(self, assign: Dict):
        "Fetches for a given assign module the submissions"
        data = {'userid': self.user_id, 'assignid': assign.get('id', 0)}
        submission = await self.client.async_post('mod_assign_get_submission_status', data)
        assign['files'] += self._get_files_of_submission(submission)

    def _get_files_of_submission(self, submission: Dict) -> List[Dict]:
        result = []
        # get own submissions
        last_attempt = submission.get('lastattempt', {})
        last_submission = last_attempt.get('submission', {})
        last_team_submission = last_attempt.get('teamsubmission', {})

        # get teachers feedback
        feedback = submission.get('feedback', {})

        result += self._get_files_of_plugins(last_submission)
        result += self._get_files_of_plugins(last_team_submission)
        result += self._get_files_of_plugins(feedback)
        result += self._get_grade_of_feedback(feedback)

        return result

    def _get_grade_of_feedback(self, feedback: Dict) -> List[Dict]:
        grade_for_display = feedback.get('gradefordisplay', "")
        graded_date = feedback.get('gradeddate', 0)
        if graded_date is None or grade_for_display is None or graded_date == 0 or grade_for_display == "":
            return []

        return [
            {
                'filename': 'grade',
                'filepath': '/',
                'timemodified': graded_date,
                'description': grade_for_display,
                'type': 'description',
            }
        ]

    def _get_files_of_plugins(self, obj: Dict) -> List[Dict]:
        result = []
        plugins = obj.get('plugins', [])

        for plugin in plugins:
            for file_area in plugin.get('fileareas', []):
                files = file_area.get('files', [])
                self.set_files_types_if_empty(files, 'submission_file')
                result += files

            for editor_field in plugin.get('editorfields', []):
                filename = editor_field.get('description', '')
                description = editor_field.get('text', '')
                if filename != '' and description != '':
                    description_file = {
                        'filename': filename,
                        'description': description,
                        'type': 'description',
                    }
                    result.append(description_file)

        return result
